# This file is part of yle-dl.
#
# Copyright 2010-2024 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from utils import fetch_title, fetch_stream_url, fetch_metadata


def test_arkisto_title():
    title = fetch_title(
        'https://yle.fi/aihe/artikkeli/2010/10/28/studio-julmahuvi-roudasta-rospuuttoon'
    )

    assert 'Roudasta rospuuttoon' in title[0]


def test_arkisto_stream_url():
    streamurls = fetch_stream_url(
        'https://yle.fi/aihe/artikkeli/2010/10/28/studio-julmahuvi-roudasta-rospuuttoon'
    )

    assert len(streamurls) >= 1
    for url in streamurls:
        assert '/manifest.mpd' in url or '.m3u8' in url


def test_arkisto_a_stream_url():
    streamurl = fetch_stream_url('https://yle.fi/aihe/a/20-10001712')

    assert len(streamurl) >= 2
    assert all(('.m3u8' in url or '.mp3' in url) for url in streamurl)


def test_arkisto_regression():
    # There was a regression (#168) related to an invalid downloadUrl
    meta = fetch_metadata(
        'https://yle.fi/aihe/artikkeli/2013/04/11/aanien-illoissa-kuunnellaan-kadonneitakin-aania'
    )
    # 12 streams in the article body, possibly more in the footer
    assert len(meta) >= 12



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import json
import itertools
import logging
import os.path
import re
from requests import HTTPError
from urllib.parse import urlparse, parse_qs


from .backends import HLSAudioBackend, DASHHLSBackend, WgetBackend
from .clip import Clip, ClipExtractor, FailedClip
from .areena_api import AreenaApiProgramInfo
from .areena_extractors import AreenaPreviewApiParser
from .kaltura import YleKalturaApiClient
from .streamflavor import StreamFlavor, failed_flavor
from .streamprobe import FullHDFlavorProber
from .timestamp import parse_areena_timestamp
from .titleformatter import TitleFormatter


logger = logging.getLogger('yledl')


def extractor_factory(url, language_chooser, httpclient, title_formatter, ffprobe):
    if (
        re.match(r'^https?://yle\.fi/aihe/', url)
        or re.match(r'^https?://svenska\.yle\.fi/artikel/', url)
        or re.match(r'^https?://svenska\.yle\.fi/a/', url)
    ):
        logger.debug(f'{url} is an Elävä Arkisto URL')
        return ElavaArkistoExtractor(
            language_chooser, httpclient, title_formatter, ffprobe
        )
    elif (
        re.match(r'^https?://areena\.yle\.fi/audio/ohjelmat/[-a-zA-Z0-9]+', url)
        or re.match(r'^https?://areena\.yle\.fi/podcastit/ohjelmat/[-a-zA-Z0-9]+', url)
        or re.match(r'^https?://areena\.yle\.fi/radio/suorat/[-a-zA-Z0-9]+', url)
    ):
        logger.debug(f'{url} is a live radio URL')
        return AreenaLiveRadioExtractor(
            language_chooser, httpclient, title_formatter, ffprobe
        )
    elif re.match(r'^https?://yle\.fi/(a|uutiset|urheilu|saa)/', url):
        logger.debug(f'{url} is a news URL')
        return YleUutisetExtractor(
            language_chooser, httpclient, title_formatter, ffprobe
        )
    elif re.match(r'^https?://(areena|arenan)\.yle\.fi/', url) or re.match(
        r'^https?://yle\.fi/', url
    ):
        logger.debug(f'{url} is an Areena URL')
        return AreenaExtractor(language_chooser, httpclient, title_formatter, ffprobe)
    elif url.lower() in ['tv1', 'tv2', 'teema']:
        logger.debug(f'{url} is a live TV channel')
        return AreenaLiveTVExtractor(
            language_chooser, httpclient, title_formatter, ffprobe
        )
    else:
        logger.debug(f'{url} is an unrecognized URL')
        return None


## Flavors


class Flavors:
    @staticmethod
    def media_type(media):
        mtype = media.get('type')
        if mtype == 'AudioObject' or (
            mtype is None and media.get('containerFormat') == 'mpeg audio'
        ):
            return 'audio'
        else:
            return 'video'


### Extract streams from an Areena webpage ###


class AreenaExtractor(ClipExtractor):
    def __init__(self, language_chooser, httpclient, title_formatter, ffprobe):
        super().__init__(httpclient)
        self.language_chooser = language_chooser
        self.title_formatter = title_formatter
        self.ffprobe = ffprobe

    def extract_clip(self, clip_url, origin_url):
        pid = self.program_id_from_url(clip_url)
        program_info = self.program_info_for_pid(
            pid, clip_url, self.title_formatter, self.ffprobe
        )
        return self.create_clip_or_failure(pid, program_info, clip_url, origin_url)

    def program_id_from_url(self, url):
        parsed = urlparse(url)
        query_dict = parse_qs(parsed.query)
        play = query_dict.get('play')
        if parsed.path.startswith('/tv/ohjelmat/') and play:
            return play[0]
        else:
            return parsed.path.split('/')[-1]

    def create_clip_or_failure(self, pid, program_info, url, origin_url):
        if not pid:
            return FailedClip(url, 'Failed to parse a program ID')

        if not program_info:
            return FailedClip(url, 'Failed to download program data', program_id=pid)

        return self.create_clip(pid, program_info, url, origin_url)

    def create_clip(self, program_id, program_info, pageurl, origin_url):
        if program_info.flavors:
            all_streams = list(
                itertools.chain.from_iterable(fl.streams for fl in program_info.flavors)
            )
        else:
            all_streams = []

        if program_info.pending:
            error_message = 'Stream not yet available.'
            if program_info.publish_timestamp:
                error_message = (
                    f'{error_message} Becomes available on '
                    f'{program_info.publish_timestamp.isoformat()}'
                )
        elif program_info.expired:
            error_message = 'This stream has expired'
        elif all_streams and all(not s.is_valid() for s in all_streams):
            error_message = all_streams[0].error_message
        elif not program_info.flavors:
            error_message = 'Media not found'
        else:
            error_message = None

        if error_message:
            return FailedClip(
                webpage=pageurl,
                error_message=error_message,
                title=program_info.title,
                description=program_info.description,
                duration_seconds=program_info.duration_seconds,
                region=program_info.available_at_region,
                publish_timestamp=program_info.publish_timestamp,
                expiration_timestamp=program_info.expiration_timestamp,
                program_id=program_id,
            )
        else:
            return Clip(
                webpage=pageurl,
                flavors=program_info.flavors,
                title=program_info.title,
                episode_title=program_info.episode_title,
                description=program_info.description,
                duration_seconds=program_info.duration_seconds,
                region=program_info.available_at_region,
                publish_timestamp=program_info.publish_timestamp,
                expiration_timestamp=program_info.expiration_timestamp,
                subtitles=program_info.subtitles,
                program_id=program_id,
                origin_url=origin_url,
                thumbnail=program_info.thumbnail,
            )

    def media_flavors(
        self,
        media_id,
        hls_manifest_url,
        download_url,
        media_type,
        is_live,
        pageurl,
        ffprobe,
    ):
        flavors = []

        if download_url:
            flavors.extend(self.download_flavors(download_url, media_type))

        flavors2 = []
        if media_id:
            flavors2.extend(
                self.flavors_by_media_id(media_id, hls_manifest_url, is_live, ffprobe)
            )

        if not flavors2 and hls_manifest_url:
            flavors2.extend(self.hls_flavors(hls_manifest_url, media_type))

        flavors.extend(flavors2)

        if self.is_kaltura_media(media_id):
            # Get mp4 streams (for wget support) from Kaltura if available.
            # Web Areena no longer uses Kaltura, so this may break (Dec 2023).
            flavors.extend(self.kaltura_mp4_flavors(media_id, pageurl))

        return flavors or None

    def flavors_by_media_id(self, media_id, hls_manifest_url, is_live, ffprobe):
        if self.is_full_hd_media(media_id) or is_live:
            logger.debug('Detected a full-HD media')
            flavors = self.hls_probe_flavors(hls_manifest_url, is_live, ffprobe)
            error = [failed_flavor('Manifest URL is missing')]
            return flavors or error
        elif self.is_html5_media(media_id):
            logger.debug('Detected an HTML5 media')
            return self.hls_probe_flavors(hls_manifest_url, False, ffprobe)
        elif self.is_media_67(media_id) or self.is_mp3_podcast(media_id):
            return []
        elif hls_manifest_url:
            # Fall-back options for new media_id types
            logger.debug('Detected a possible HLS media')
            return self.hls_probe_flavors(hls_manifest_url, False, ffprobe)
        else:
            return [failed_flavor('Unknown stream flavor')]

    def kaltura_mp4_flavors(self, media_id, pageurl):
        entry_id = self.kaltura_entry_id(media_id)
        kapi_client = YleKalturaApiClient(self.httpclient)
        playback_context = kapi_client.playback_context(entry_id, pageurl)
        if playback_context:
            return kapi_client.parse_stream_flavors(playback_context, pageurl)
        else:
            return []

    def is_html5_media(self, media_id):
        # 29- is the most common media ID
        # 84-, hosted on yleawsmpondemand-04.akamaized.net, April 2024
        # 85-, ylekvodmod01.akamaized.net, also seen on podcasts, Summer 2024
        return media_id and (
            media_id.startswith('29-')
            or media_id.startswith('84-')
            or media_id.startswith('85-')
        )

    def is_kaltura_media(self, media_id):
        return media_id and media_id.startswith('29-')

    def is_full_hd_media(self, media_id):
        return media_id and media_id.startswith('55-')

    def is_media_67(self, media_id):
        # A new hosting alternative (June 2021)? Hosted on yleawsmpodamdipv4.akamaized.net
        return media_id and media_id.startswith('67-')

    def is_mp3_podcast(self, media_id):
        # Podcast streams, "78-" seen on Spring 2023
        # Prefer download_url, no extra flavors here.
        return media_id and media_id.startswith('78-')

    def is_live_media(self, media_id):
        return media_id and media_id.startswith('10-')

    def kaltura_entry_id(self, mediaid):
        return mediaid.split('-', 1)[-1]

    def hls_flavors(self, hls_manifest_url, media_type):
        if not hls_manifest_url:
            return []

        if media_type == 'video':
            backend = DASHHLSBackend(hls_manifest_url, experimental_subtitles=True)
        else:
            backend = HLSAudioBackend(hls_manifest_url)

        return [StreamFlavor(media_type=media_type, streams=[backend])]

    def hls_probe_flavors(self, hls_manifest_url, is_live, ffprobe):
        if not hls_manifest_url:
            return []

        logger.debug('Probing for stream flavors')
        return FullHDFlavorProber().probe_flavors(hls_manifest_url, is_live, ffprobe)

    def download_flavors(self, download_url, media_type):
        path = urlparse(download_url)[2]
        ext = os.path.splitext(path)[1] or None
        backend = WgetBackend(download_url, ext)
        return [StreamFlavor(media_type=media_type, streams=[backend])]

    def publish_event(self, program_info):
        events = (program_info or {}).get('data', {}).get('publicationEvent', [])
        areena_events = [
            e for e in events if e.get('service', {}).get('id') == 'yle-areena'
        ]
        has_current = any(self.publish_event_is_current(e) for e in areena_events)
        if has_current:
            areena_events = [
                e for e in areena_events if self.publish_event_is_current(e)
            ]

        with_media = [e for e in areena_events if e.get('media')]
        if with_media:
            sorted_events = sorted(
                with_media, key=lambda e: e.get('startTime'), reverse=True
            )
            return sorted_events[0]
        else:
            return {}

    def publish_timestamp(self, program_info):
        ts = self.publish_event(program_info).get('startTime')
        return parse_areena_timestamp(ts)

    def program_info_for_pid(self, pid, pageurl, title_formatter, ffprobe):
        if not pid:
            return None

        preview = self.preview_parser(pid, pageurl)
        publish_timestamp = preview.timestamp()
        titles = preview.title(self.language_chooser)
        title_params = {
            'title': '',
            'program_id': pid,
            'publish_timestamp': publish_timestamp,
        }
        title_params.update(titles)
        season_and_episode = preview.season_and_episode()
        if season_and_episode and 'season' not in season_and_episode:
            logger.debug('Checking the webpage for a season number')
            season_and_episode.update(self.extract_season_number(pageurl))
        title_params.update(season_and_episode)
        title = title_formatter.format(**title_params) or 'areena'
        simple_formatter = TitleFormatter('${series_separator}${title}')
        episode_title = simple_formatter.format(**title_params)
        media_id = preview.media_id()
        is_live = self.is_live_media(media_id) or preview.is_live()
        download_url = self.ignore_invalid_download_url(preview.media_url())
        if self.is_html5_media(media_id):
            preview_subtitles = preview.subtitles()
        else:
            preview_subtitles = []

        return AreenaApiProgramInfo(
            media_id=media_id,
            title=title,
            episode_title=episode_title,
            description=preview.description(self.language_chooser),
            flavors=self.media_flavors(
                media_id,
                preview.manifest_url(),
                download_url,
                preview.media_type(),
                is_live,
                pageurl,
                ffprobe,
            ),
            thumbnail=preview.thumbnail_url(),
            subtitles=preview_subtitles,
            duration_seconds=preview.duration_seconds(),
            available_at_region=preview.available_at_region() or 'Finland',
            publish_timestamp=publish_timestamp,
            expiration_timestamp=None,
            pending=preview.is_pending(),
            expired=preview.is_expired(),
        )

    def preview_parser(self, pid, pageurl):
        preview_headers = {'Referer': pageurl, 'Origin': 'https://areena.yle.fi'}
        url = self.preview_url(pid)
        try:
            preview_json = self.httpclient.download_json(url, preview_headers)
        except HTTPError as ex:
            if ex.response.status_code == 404:
                logger.warning(f'Preview API result not found: {url}')
                preview_json = []
            else:
                raise
        logger.debug(f'preview data:\n{json.dumps(preview_json, indent=2)}')

        return AreenaPreviewApiParser(preview_json)

    def preview_url(self, program_id):
        return (
            f'https://player.api.yle.fi/v1/preview/{program_id}.json?'
            'language=fin'
            '&ssl=true'
            '&countryCode=FI'
            '&host=areenaylefi'
            '&app_id=player_static_prod'
            '&app_key=8930d72170e48303cf5f3867780d549b'
            '&isPortabilityRegion=true'
        )

    def publish_event_is_current(self, event):
        return event.get('temporalStatus') == 'currently'

    def ignore_invalid_download_url(self, url):
        # Sometimes download url is missing the file name
        return None if (url and url.endswith('/')) else url

    def extract_season_number(self, pageurl):
        # TODO: how to get the season number without downloading the HTML page?
        tree = self.httpclient.download_html_tree(pageurl)
        title_tag = tree.xpath('/html/head/title/text()')
        if len(title_tag) > 0:
            title = title_tag[0]
            m = re.match(r'K(\d+), J\d+', title)
            if m:
                return {'season': int(m.group(1))}

        return {}


### Areena live TV ###


class AreenaLiveTVExtractor(AreenaExtractor):
    def get_playlist(self, url, latest_only=False):
        return [url]

    def program_id_from_url(self, url):
        known_channels = {
            'tv1': 'yle-tv1',
            'tv2': 'yle-tv2',
            'teema': 'yle-teema-fem',
        }

        return known_channels.get(url.lower())


### Areena live radio ###


class AreenaLiveRadioExtractor(AreenaExtractor):
    def get_playlist(self, url, latest_only=False):
        return [url]

    def program_id_from_url(self, url):
        known_channels = {
            '57-p89RepWE0': 'yle-radio-1',
            '57-JAprnp7W2': 'ylex',
            '57-kpDBBz8Pz': 'yle-puhe',
            '57-md5vJP6a2': 'yle-x3m',
            '57-llL6Y4blL': 'yle-klassinen',
            '57-bN8gjw7AY': 'yle-sami-radio',
            # Radio Suomi and Vega have regional channels selected by the query
            # parameter _c. If _c is missing, use these hard coded values.
            '57-3gO4bl7J6': 'yle-radio-suomi-helsinki',
            '57-P3mO0mdm6': 'radio-vega-huvudstadsregionen',
        }

        parsed = urlparse(url)
        query_dict = parse_qs(parsed.query)
        if query_dict.get('_c'):
            return query_dict.get('_c')[0]
        else:
            key = parsed.path.split('/')[-1]
            return known_channels.get(key, key)


### Elava Arkisto ###


class ElavaArkistoExtractor(AreenaExtractor):
    def get_playlist(self, url, latest_only=False):
        ids = self.get_dataids(url)

        if latest_only:
            ids = ids[-1:]

        if ids:
            return [f'https://areena.yle.fi/{x}' for x in ids]
        else:
            # Fallback to Yle news parser because sometimes Elävä
            # arkisto pages are published using the same article type
            # as news articles.
            return parse_playlist_from_yle_article(url, self.httpclient, latest_only)

    def get_dataids(self, url):
        tree = self.httpclient.download_html_tree(url)
        if tree is None:
            return []

        return self.ordered_union(self._simple_dataids(tree), self._ydd_dataids(tree))

    def ordered_union(self, xs, ys):
        union = list(xs)  # copy
        for y in ys:
            if y not in union:
                union.append(y)
        return union

    def _simple_dataids(self, tree):
        dataids = tree.xpath("//article[@id='main-content']//div/@data-id")
        dataids = [str(d) for d in dataids]
        return [d if '-' in d else f'1-{d}' for d in dataids]

    def _ydd_dataids(self, tree):
        player_props = [
            json.loads(p)
            for p in tree.xpath("//main[@id='main-content']//div/@data-player-props")
        ]
        return [x['id'] for x in player_props if 'id' in x]


### News clips at the Yle news site ###


class YleUutisetExtractor(AreenaExtractor):
    def get_playlist(self, url, latest_only=False):
        return parse_playlist_from_yle_article(url, self.httpclient, latest_only)


def parse_playlist_from_yle_article(url, httpclient, latest_only):
    def id_to_areena_url(data_id):
        if '-' in data_id:
            areena_id = data_id
        else:
            areena_id = f'1-{data_id}'
        return f'https://areena.yle.fi/{areena_id}'

    tree = httpclient.download_html_tree(url)
    if tree is None:
        return []

    state = None
    state_script_nodes = tree.xpath(
        '//script[@type="text/javascript" and '
        '(contains(text(), "window.__INITIAL__STATE__") or '
        ' contains(text(), "window.__INITIAL_STATE__"))]/text()'
    )
    if len(state_script_nodes) > 0:
        state_json = re.sub(
            r'^window\.__INITIAL__?STATE__\s*=\s*', '', state_script_nodes[0]
        )
        state = json.loads(state_json)

    if state is None:
        state_div_nodes = tree.xpath('//div[@id="initialState"]')
        if len(state_div_nodes) > 0:
            state = json.loads(state_div_nodes[0].attrib.get('data-state'))

    if state is None:
        return []

    data_ids = []
    article = state.get('pageData', {}).get('article', {})
    if article.get('mainMedia') is not None:
        medias = article['mainMedia']
        data_ids = [
            media['id']
            for media in medias
            if media.get('type') in ['VideoBlock', 'video'] and 'id' in media
        ]
    else:
        headline_video_id = article.get('headline', {}).get('video', {}).get('id')
        if headline_video_id:
            data_ids = [headline_video_id]

    content = article.get('content', [])
    inline_media = [
        block['id']
        for block in content
        if block.get('type') in ['AudioBlock', 'audio', 'VideoBlock', 'video']
        and 'id' in block
    ]
    for id in inline_media:
        if id not in data_ids:
            data_ids.append(id)

    logger.debug(f'Found Areena data IDs: {",".join(data_ids)}')

    playlist = [id_to_areena_url(id) for id in data_ids]
    if latest_only:
        playlist = playlist[-1:]

    return playlist



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from utils import fetch_metadata


def test_uutiset_main_media_metadata():
    # This page has a video embedded as "mainMedia"
    metadata = fetch_metadata('https://yle.fi/a/3-11822823')

    assert len(metadata) == 1
    assert metadata[0]['webpage'] == 'https://areena.yle.fi/1-50653021'
    assert metadata[0]['title'].startswith(
        'Perjantai-dokkari: Tuomas, vaatimaton maailmantähti'
    )
    assert metadata[0]['duration_seconds'] == 690
    assert metadata[0]['region'] == 'World'
    assert metadata[0]['publish_timestamp'] == '2021-03-05T21:30:00+02:00'
    assert 'expired_timestamp' not in metadata[0]
    assert len(metadata[0]['description']) > 150

    flavors = metadata[0]['flavors']
    assert len(flavors) >= 3
    assert any(f.get('media_type') == 'video' for f in flavors)
    assert all('bitrate' in f for f in flavors)
    assert any('height' in f and 'width' in f for f in flavors)


def test_uutiset_headline_metadata():
    # This page has a video embedded as "headline.video"
    metadata = fetch_metadata('https://yle.fi/a/3-12328632')

    assert len(metadata) == 1
    assert metadata[0]['webpage'] == 'https://areena.yle.fi/1-61842917'
    assert metadata[0]['title'].startswith(
        'Uutisvideot 2022: Presidentti Sauli Niinistö'
    )
    assert metadata[0]['duration_seconds'] == 49
    assert metadata[0]['region'] == 'World'
    assert metadata[0]['publish_timestamp'] == '2022-02-22T14:36:36+02:00'
    assert 'expired_timestamp' not in metadata[0]
    assert len(metadata[0]['description']) > 100

    flavors = metadata[0]['flavors']
    assert len(flavors) >= 1
    assert all(f.get('media_type') == 'video' for f in flavors)


def test_uutiset_inline_video_block_metadata():
    # This page has two videos embedded as inline blocks
    metadata = fetch_metadata('https://yle.fi/a/74-20036911')

    assert len(metadata) == 2
    assert metadata[0]['title'].startswith(
        'Minuutin uutisvideot: Kymmenien ihmisten ryhmissä kuljetaan pummilla'
    )
    assert metadata[0]['duration_seconds'] == 49
    assert 'Raitiovaunuliikennettä Helsingin Hakaniemessä' in metadata[1]['title']
    assert metadata[1]['duration_seconds'] == 14

    assert any(f.get('media_type') == 'video' for f in metadata[0]['flavors'])
    assert any(f.get('media_type') == 'video' for f in metadata[1]['flavors'])


def test_uutiset_inline_audio_block_metadata():
    # This page has an inline AudioBlock
    metadata = fetch_metadata('https://yle.fi/a/3-11843834')

    assert len(metadata) == 1
    assert metadata[0]['webpage'] == 'https://areena.yle.fi/1-50762351'
    assert metadata[0]['title'].startswith(
        'Ykkösaamun kolumni: Janne Saarikivi: On kriisi'
    )
    assert metadata[0]['duration_seconds'] == 333
    assert metadata[0]['region'] == 'World'
    assert metadata[0]['publish_timestamp'] == '2021-03-23T07:00:00+02:00'
    assert 'expired_timestamp' not in metadata[0]
    assert len(metadata[0]['description']) > 150

    flavors = metadata[0]['flavors']
    assert len(flavors) >= 1
    assert all(f.get('media_type') == 'audio' for f in flavors)


def test_uutiset_svenska():
    metadata = fetch_metadata('https://svenska.yle.fi/a/7-10023932')

    assert len(metadata) >= 1
    assert metadata[0]['webpage'] == 'https://areena.yle.fi/1-64379365'
    assert metadata[0]['title'].startswith('Yle Nyheter')
    assert metadata[0]['duration_seconds'] == 73
    assert metadata[0]['region'] == 'World'
    assert metadata[0]['publish_timestamp'] == '2022-12-05T12:28:13+02:00'
    assert 'expired_timestamp' not in metadata[0]

    flavors = metadata[0]['flavors']
    assert len(flavors) >= 1
    assert all(f.get('media_type') == 'video' for f in flavors)


def test_uutiset_metadata_old_address():
    # Before Nov 2022, news articles were published at /uutiset/
    # instead of /a/
    metadata = fetch_metadata('https://yle.fi/uutiset/74-20006891')

    assert len(metadata) == 1
    assert metadata[0]['webpage'] == 'https://areena.yle.fi/1-64358035'
    assert metadata[0]['title'].startswith(
        'Puoli seitsemän: Neljä pukeutumisvinkkiä talvipyöräilyyn'
    )
    assert metadata[0]['duration_seconds'] == 74
    assert metadata[0]['region'] == 'World'
    assert metadata[0]['publish_timestamp'] == '2022-12-02T10:09:44+02:00'
    assert 'expired_timestamp' not in metadata[0]

    flavors = metadata[0]['flavors']
    assert len(flavors) >= 1
    assert all(f.get('media_type') == 'video' for f in flavors)



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import sys
import json
from datetime import timedelta, tzinfo
from io import BytesIO
from typing import Tuple
from yledl import execute_action, StreamFilters, IOContext, StreamAction, RD_SUCCESS
from yledl.io import random_elisa_ipv4
from yledl.http import HttpClient
from yledl.titleformatter import TitleFormatter


# Context manager for capturing stdout output. See
# https://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call
class Capturing(list):
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._bytesio = BytesIO()
        return self

    def __exit__(self, *args):
        self.extend(self._bytesio.getvalue().decode('UTF-8').splitlines())
        del self._bytesio  # free up some memory
        sys.stdout = self._stdout


class FixedOffset(tzinfo):
    def __init__(self, offset_hours):
        self.__offset = timedelta(hours=offset_hours)

    def utcoffset(self, dt):
        return self.__offset

    def tzname(self, dt):
        return 'FixedOffset'

    def dst(self, dt):
        return timedelta(0)


class MockIOContext(IOContext):
    def ffmpeg_version(self) -> Tuple[int, int]:
        print('Mock ffmpeg_version called!')
        return 7, 1


def fetch_title(url, filters=StreamFilters()):
    return fetch(url, StreamAction.PRINT_STREAM_TITLE, filters)


def fetch_stream_url(url, filters=StreamFilters()):
    return fetch(url, StreamAction.PRINT_STREAM_URL, filters)


def fetch_episode_pages(url, filters=StreamFilters()):
    return fetch(url, StreamAction.PRINT_EPISODE_PAGES, filters)


def fetch_metadata(url, filters=StreamFilters(), meta_language=None):
    return json.loads(
        '\n'.join(fetch(url, StreamAction.PRINT_METADATA, filters, meta_language))
    )


def fetch(url, action, filters, meta_language=None):
    io = MockIOContext(
        destdir='/tmp/',
        metadata_language=meta_language,
        x_forwarded_for=random_elisa_ipv4(),
    )
    httpclient = HttpClient(io)
    title_formatter = TitleFormatter()

    with Capturing() as output:
        res = execute_action(
            url, action, io, httpclient, title_formatter, stream_filters=filters
        )
        assert res == RD_SUCCESS

    return output



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional
from datetime import datetime

from dataclasses import dataclass
from .streamflavor import StreamFlavor
from .subtitles import Subtitle


@dataclass(frozen=True)
class AreenaApiProgramInfo:
    media_id: str
    title: str
    episode_title: str
    description: Optional[str]
    flavors: List[StreamFlavor]
    thumbnail: Optional[str]
    subtitles: List[Subtitle]
    duration_seconds: Optional[int]
    available_at_region: str
    publish_timestamp: Optional[datetime]
    expiration_timestamp: Optional[datetime]
    pending: bool
    expired: bool


@dataclass(frozen=True)
class EpisodeMetadata:
    uri: str
    season_number: Optional[int]
    episode_number: Optional[int]
    release_date: Optional[datetime]

    def sort_key(self):
        return (
            self.season_number or 99999,
            self.episode_number or 99999,
            self.release_date or datetime(1970, 1, 1, 0, 0, 0),
        )

    def with_episode_number(self, ep):
        return EpisodeMetadata(self.uri, self.season_number, ep, self.release_date)



# This file is part of yle-dl.
#
# Copyright 2010-2024 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, List

from dataclasses import dataclass
from .http import update_url_query


@dataclass(frozen=True)
class PlaylistData:
    # The base URL from which to download a playlist
    base_url: str
    # List of query parameters. Each item is a dictionary of query
    # parameters for one season. If empty, a playlist is downloaded
    # from the plain base_url.
    season_parameters: List[Dict]

    def season_playlist_urls(self):
        if self.season_parameters:
            for season_query in self.season_parameters:
                yield update_url_query(self.base_url, season_query)
        else:
            yield self.base_url



# This file is part of yle-dl.
#
# Copyright 2010-2022 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import pytest


def pytest_addoption(parser):
    parser.addoption(
        '--geoblocked',
        action='store_true',
        help='Enable get-blocked tests that work only in Finland',
    )


def pytest_configure(config):
    config.addinivalue_line(
        'markers', 'geoblocked: get-blocked test that work only in Finland'
    )


def pytest_collection_modifyitems(config, items):
    if not config.option.geoblocked:
        # Skip tests marked as geoblocked
        skip_geoblocked = pytest.mark.skip(reason='need --geoblocked option to run')
        for item in items:
            if 'geoblocked' in item.keywords:
                item.add_marker(skip_geoblocked)



# This file is part of yle-dl.
#
# Copyright 2010-2024 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import pytest
from datetime import datetime
from utils import fetch_title, fetch_stream_url, fetch_metadata
from yledl import StreamFilters


def test_radio_title():
    title = fetch_title('https://areena.yle.fi/1-3361013')

    assert len(title) == 1
    assert title[0].startswith(
        'Tiedeykkönen: Suorat aurinkobiopolttoaineet mullistavat energiantuotannon'
    )


def test_radio_stream_url_hls():
    url = fetch_stream_url('https://areena.yle.fi/1-4551633')

    assert len(url) == 1
    assert '.mp3' in url[0]


def test_radio_metadata_hls():
    metadata = fetch_metadata('https://areena.yle.fi/1-4551633')

    assert len(metadata) == 1
    assert len(metadata[0]['flavors']) >= 1
    assert metadata[0]['flavors'][0]['media_type'] == 'audio'
    assert metadata[0]['duration_seconds'] == 2954
    assert len(metadata[0]['description']) > 150


@pytest.mark.geoblocked
def test_radio_live_url():
    url = fetch_stream_url('https://areena.yle.fi/podcastit/ohjelmat/57-kpDBBz8Pz')

    assert len(url) == 1
    assert '.m3u8' in url[0]


@pytest.mark.geoblocked
def test_radio_live_url2():
    url = fetch_stream_url(
        'https://areena.yle.fi/podcastit/ohjelmat/57-3gO4bl7J6?_c=yle-radio-suomi-oulu'
    )

    assert len(url) == 1
    assert '.m3u8' in url[0]


@pytest.mark.geoblocked
def test_radio_live_metadata():
    metadata = fetch_metadata('https://areena.yle.fi/podcastit/ohjelmat/57-kpDBBz8Pz')

    assert len(metadata) == 1
    assert len(metadata[0]['flavors']) >= 1
    assert all(f.get('media_type') == 'audio' for f in metadata[0]['flavors'])
    assert metadata[0]['title'].startswith('Yle Puhe')


def test_radio_series_2020():
    urls = fetch_stream_url('https://areena.yle.fi/podcastit/1-50198109')

    assert len(urls) >= 6


def test_radio_series_redirect():
    # Will get redirected to https://areena.yle.fi/podcastit/1-61070264
    urls = fetch_stream_url('https://areena.yle.fi/1-61070264')

    assert len(urls) >= 10


def test_radio_series_redirect_from_old_audio_url():
    # The address for radio programs was changed from /audio/ to /podcastit/
    # in September 2022. Test that the old address is redirected to the new.
    urls = fetch_stream_url('https://areena.yle.fi/audio/1-61070277')

    assert len(urls) == 1
    assert '.mp3' in urls[0]


def test_radio_metadata_2020():
    metadata = fetch_metadata('https://areena.yle.fi/podcastit/1-50198110')

    assert len(metadata) == 1
    assert len(metadata[0]['flavors']) >= 1
    assert metadata[0]['flavors'][0]['media_type'] == 'audio'
    assert metadata[0]['duration_seconds'] == 1451
    assert len(metadata[0]['description']) > 150


def test_radio_metadata_media_id_78():
    # This has a media_id starting with 78-
    metadata = fetch_metadata('https://areena.yle.fi/podcastit/1-65772446')

    assert len(metadata) == 1
    assert len(metadata[0]['flavors']) >= 1
    assert not any('error' in fl for fl in metadata[0]['flavors'])
    assert metadata[0]['flavors'][0]['media_type'] == 'audio'
    assert metadata[0]['duration_seconds'] == 1769
    assert len(metadata[0]['description']) > 150


def test_radio_episodes_sort_order_latest_last_source():
    # This page lists episodes in the latest-last order
    metadata = fetch_metadata('https://areena.yle.fi/podcastit/1-50375734')

    # Should be sorted from oldest to newest
    timestamps = [x['publish_timestamp'] for x in metadata]
    assert len(timestamps) > 1
    assert timestamps == sorted(timestamps)


def test_radio_latest():
    # Test fetching the latest radio episode.
    filters = StreamFilters(latest_only=True)
    metadata = fetch_metadata('https://areena.yle.fi/podcastit/1-3529062', filters)

    assert len(metadata) == 1

    publish_date = datetime.strptime(metadata[0]['publish_timestamp'][:10], '%Y-%m-%d')
    assert publish_date >= datetime(2016, 7, 28)


def test_radio_return_empty_list_if_there_are_no_episodes():
    # This series has no published episodes
    metadata = fetch_metadata('https://areena.yle.fi/podcastit/1-3114806')

    assert len(metadata) == 0



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import copy
import logging
import unittest.mock
import pytest
from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime
from unittest.mock import Mock
from utils import FixedOffset, MockIOContext
from yledl import StreamFilters, RD_SUCCESS, RD_FAILED
from yledl.backends import BaseDownloader, FailingBackend
from yledl.downloader import YleDlDownloader
from yledl.errors import TransientDownloadError
from yledl.clip import Clip, FailedClip
from yledl.extractors import StreamFlavor
from yledl.titleformatter import TitleFormatter


class MockGeoLocation:
    def located_in_finland(self, referrer):
        return True


class MockExtractor:
    def __init__(self, clips_by_url):
        self.clips_by_url = clips_by_url
        self.title_formatter = TitleFormatter()

    def extract(self, url, latest_only):
        return list(self.clips_by_url.values())

    def get_playlist(self, url, latest_only=False):
        return list(self.clips_by_url.keys())

    def extract_clip(self, url, origin_url):
        return self.clips_by_url[url]


def mock_backend(
    status=RD_SUCCESS,
    name='ffmpeg',
    stream_url='https://areena.example.com/video/areena.mp4',
):
    backend = BaseDownloader()
    backend.name = name
    backend.save_stream = Mock(return_value=status)
    backend.pipe = Mock(return_value=status)
    backend.stream_url = Mock(return_value=stream_url)
    return backend


def backend_that_fails_n_times(n):
    """Return a BaseDownloader instance that first fails and then succeeds.

    The first n calls to save_stream() and pipe() return RD_FAILED and the next
     call after that return RD_SUCCESS.

    save_stream() and pipe() are Mock instances.
    """
    backend = BaseDownloader()
    backend.name = 'ffmpeg'
    return_values = [TransientDownloadError('Failed!')] * n + [RD_SUCCESS]
    backend.save_stream = Mock(side_effect=return_values)
    backend.pipe = Mock(side_effect=return_values)
    return backend


def successful_clip(title='Test clip: S01E01-2018-07-01T00:00'):
    flavors = [
        # The flavors are intentionally unsorted
        StreamFlavor(
            media_type='video',
            height=1080,
            width=1920,
            bitrate=2808,
            streams=[
                mock_backend(stream_url='https://example.com/video/high_quality.mp4')
            ],
        ),
        StreamFlavor(
            media_type='video',
            height=360,
            width=640,
            bitrate=880,
            streams=[
                mock_backend(stream_url='https://example.com/video/low_quality.mp4')
            ],
        ),
        StreamFlavor(
            media_type='video',
            height=480,
            width=640,
            bitrate=964,
            streams=[
                mock_backend(stream_url='https://example.com/video/low_quality_2.mp4')
            ],
        ),
        StreamFlavor(
            media_type='video',
            height=720,
            width=1280,
            bitrate=1412,
            streams=[
                mock_backend(stream_url='https://example.com/video/medium_quality.mp4')
            ],
        ),
        StreamFlavor(
            media_type='video',
            height=720,
            width=1280,
            bitrate=1872,
            streams=[
                mock_backend(
                    stream_url='https://example.com/video/medium_quality_high_bitrate.mp4'
                )
            ],
        ),
    ]
    return create_clip(flavors, title)


def incomplete_flavors_clip():
    return Clip(
        webpage='https://areena.yle.fi/1-1234567',
        flavors=[
            StreamFlavor(
                media_type='video',
                streams=[mock_backend(stream_url='https://example.com/video/1.mp4')],
            ),
            StreamFlavor(
                media_type='video',
                height=360,
                width=640,
                streams=[mock_backend(stream_url='https://example.com/video/2.mp4')],
            ),
            StreamFlavor(
                media_type='video',
                streams=[mock_backend(stream_url='https://example.com/video/3.mp4')],
            ),
        ],
        title='Test clip: S01E01-2018-07-01T00:00',
        duration_seconds=None,
        region='Finland',
        publish_timestamp=None,
        expiration_timestamp=None,
    )


def multistream_clip():
    return create_clip(
        [
            StreamFlavor(
                media_type='video',
                height=360,
                width=640,
                bitrate=964,
                streams=[
                    FailingBackend('Invalid stream'),
                    FailingBackend('Invalid stream'),
                    mock_backend(
                        name='wget', stream_url='https://example.com/video/3.mp4'
                    ),
                    mock_backend(
                        name='ffmpeg', stream_url='https://example.com/video/4.mp4'
                    ),
                ],
            )
        ]
    )


def failed_clip():
    return FailedClip('https://areena.yle.fi/1-1234567', 'Failed test clip')


def failed_stream_clip():
    return create_clip(
        [
            StreamFlavor(
                media_type='video',
                height=360,
                width=640,
                bitrate=964,
                streams=[
                    FailingBackend('Invalid stream'),
                    FailingBackend('Invalid stream'),
                ],
            )
        ]
    )


def create_clip(flavors, title='Test clip: S01E01-2018-07-01T00:00'):
    return Clip(
        webpage='https://areena.yle.fi/1-1234567',
        flavors=flavors,
        title=title,
        duration_seconds=950,
        region='Finland',
        publish_timestamp=datetime(2018, 7, 1, tzinfo=FixedOffset(3)),
        expiration_timestamp=datetime(2019, 1, 1, tzinfo=FixedOffset(3)),
    )


@dataclass(frozen=True)
class DownloaderParametersFixture:
    io: MockIOContext
    filters: StreamFilters


@pytest.fixture
def simple():
    return DownloaderParametersFixture(
        io=MockIOContext(destdir='/tmp/'), filters=StreamFilters()
    )


def downloader(clips_by_url):
    def extractor_factory(*args):
        return MockExtractor(clips_by_url)

    return YleDlDownloader(MockGeoLocation(), TitleFormatter(), None, extractor_factory)


def stream_by_partial_url_match(clip, url_contains):
    for flavor in clip.flavors:
        for stream in flavor.streams:
            if url_contains in stream.stream_url():
                return stream

    return None


def test_download_success(simple):
    clip = successful_clip()
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, simple.filters)

    clip.flavors[0].streams[0].save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_download_incomplete_metadata(simple):
    clip = incomplete_flavors_clip()
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, simple.filters)

    # flavors[1] is expected to be downloaded because it is the only flavor that
    # has height and width metadata
    clip.flavors[0].streams[0].save_stream.assert_not_called()
    clip.flavors[1].streams[0].save_stream.assert_called_once()
    clip.flavors[2].streams[0].save_stream.assert_not_called()
    assert res == RD_SUCCESS


def test_download_filter_resolution(simple):
    clip = successful_clip()
    filters = StreamFilters(maxheight=400)
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, filters)

    stream_by_partial_url_match(clip, 'low_quality').save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_download_filter_exact_resolution(simple):
    clip = successful_clip()
    filters = StreamFilters(maxheight=720)
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, filters)

    stream_by_partial_url_match(clip, 'medium_quality').save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_download_filter_bitrate1(simple):
    clip = successful_clip()
    filters = StreamFilters(maxbitrate=1500)
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, filters)

    stream_by_partial_url_match(clip, 'medium_quality').save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_download_filter_bitrate2(simple):
    clip = successful_clip()
    filters = StreamFilters(maxbitrate=2000)
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, filters)

    stream_by_partial_url_match(
        clip, 'medium_quality_high_bitrate'
    ).save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_download_multiple_filters(simple):
    clip = successful_clip()
    filters = StreamFilters(maxheight=700, maxbitrate=900)
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, filters)

    stream_by_partial_url_match(clip, 'low_quality').save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_pipe_success(simple):
    clip = successful_clip()
    dl = downloader({'a': clip})
    res = dl.pipe('', simple.io, simple.filters)

    stream_by_partial_url_match(clip, 'high_quality').pipe.assert_called_once()
    assert res == RD_SUCCESS


def test_print_urls(simple):
    dl = downloader(
        OrderedDict(
            [
                ('a', successful_clip()),
                ('b', successful_clip()),
            ]
        )
    )
    urls = list(dl.get_urls('', simple.io, simple.filters))

    assert urls == [
        'https://example.com/video/high_quality.mp4',
        'https://example.com/video/high_quality.mp4',
    ]


def test_print_titles(simple):
    titles = ['Uutiset', 'Pasila: S01E01-2018-07-01T00:00']
    dl = downloader(
        OrderedDict(
            [
                ('a', successful_clip(titles[0])),
                ('b', successful_clip(titles[1])),
            ]
        )
    )

    assert list(dl.get_titles('', simple.io, False)) == titles


def test_print_titles_replaces_whitespace(simple):
    titles = ['   Title with\tall\vkinds\u00a0of\u2003whitespace \t \u00a0 characters']
    expected_titles = ['Title with all kinds of whitespace characters']
    dl = downloader({'a': successful_clip(titles[0])})

    assert list(dl.get_titles('', simple.io, False)) == expected_titles


def test_print_metadata(simple):
    dl = downloader({'a': successful_clip()})
    metadata = dl.get_metadata('', simple.io, False)

    assert len(metadata) == 1

    # Match filename fuzzily because the exact name depends on the existing
    # file names
    assert 'Test clip' in metadata[0]['filename']
    del metadata[0]['filename']
    assert metadata == [
        {
            'webpage': 'https://areena.yle.fi/1-1234567',
            'title': 'Test clip: S01E01-2018-07-01T00:00',
            'episode_title': '',
            'flavors': [
                {
                    'media_type': 'video',
                    'height': 360,
                    'width': 640,
                    'bitrate': 880,
                    'url': 'https://example.com/video/low_quality.mp4',
                    'backends': ['ffmpeg'],
                },
                {
                    'media_type': 'video',
                    'height': 480,
                    'width': 640,
                    'bitrate': 964,
                    'url': 'https://example.com/video/low_quality_2.mp4',
                    'backends': ['ffmpeg'],
                },
                {
                    'media_type': 'video',
                    'height': 720,
                    'width': 1280,
                    'bitrate': 1412,
                    'url': 'https://example.com/video/medium_quality.mp4',
                    'backends': ['ffmpeg'],
                },
                {
                    'media_type': 'video',
                    'height': 720,
                    'width': 1280,
                    'bitrate': 1872,
                    'url': 'https://example.com/video/medium_quality_high_bitrate.mp4',
                    'backends': ['ffmpeg'],
                },
                {
                    'media_type': 'video',
                    'height': 1080,
                    'width': 1920,
                    'bitrate': 2808,
                    'url': 'https://example.com/video/high_quality.mp4',
                    'backends': ['ffmpeg'],
                },
            ],
            'duration_seconds': 950,
            'subtitles': [],
            'region': 'Finland',
            'publish_timestamp': '2018-07-01T00:00:00+03:00',
            'expiration_timestamp': '2019-01-01T00:00:00+03:00',
        }
    ]


def test_print_metadata_incomplete(simple):
    dl = downloader({'a': incomplete_flavors_clip()})
    metadata = dl.get_metadata('', simple.io, False)

    assert len(metadata) == 1

    # Match filename fuzzily because the exact name depends on the existing
    # file names
    assert 'Test clip' in metadata[0]['filename']
    del metadata[0]['filename']

    assert metadata == [
        {
            'webpage': 'https://areena.yle.fi/1-1234567',
            'title': 'Test clip: S01E01-2018-07-01T00:00',
            'episode_title': '',
            'flavors': [
                {
                    'media_type': 'video',
                    'url': 'https://example.com/video/1.mp4',
                    'backends': ['ffmpeg'],
                },
                {
                    'media_type': 'video',
                    'height': 360,
                    'width': 640,
                    'url': 'https://example.com/video/2.mp4',
                    'backends': ['ffmpeg'],
                },
                {
                    'media_type': 'video',
                    'url': 'https://example.com/video/3.mp4',
                    'backends': ['ffmpeg'],
                },
            ],
            'region': 'Finland',
            'subtitles': [],
        }
    ]


def test_download_failed_clip(simple):
    dl = downloader({'a': failed_clip()})
    res = dl.download_clips('', simple.io, simple.filters)

    assert res == RD_FAILED


def test_download_failed_stream(simple):
    dl = downloader({'a': failed_stream_clip()})
    res = dl.download_clips('', simple.io, simple.filters)

    assert res == RD_FAILED


def test_print_metadata_failed_clip(simple):
    dl = downloader({'a': failed_clip()})
    metadata = dl.get_metadata('', simple.io, False)

    assert metadata == [
        {
            'webpage': 'https://areena.yle.fi/1-1234567',
            'flavors': [{'error': failed_clip().flavors[0].streams[0].error_message}],
            'region': 'Finland',
            'title': '',
            'episode_title': '',
            'subtitles': [],
        }
    ]


def test_download_fallback(simple):
    clip = multistream_clip()
    dl = downloader({'a': clip})
    res = dl.download_clips('', simple.io, simple.filters)

    clip.flavors[0].streams[3].save_stream.assert_called_once()
    assert res == RD_SUCCESS


def test_postprocessing_no_log_errors(simple):
    # Smoke test for PR #303
    dl = downloader({'a': successful_clip()})
    io_postprocess = copy.copy(simple.io)
    io_postprocess.postprocess_command = 'echo'
    logger = logging.getLogger('yledl')
    with unittest.mock.patch.object(logger, 'error') as mock_log_error:
        res = dl.download_clips('', io_postprocess, simple.filters)

        mock_log_error.assert_not_called()

    assert res == RD_SUCCESS


def test_download_successful_after_retry(simple):
    backend = backend_that_fails_n_times(2)
    flavors = [
        StreamFlavor(
            media_type='video', height=1080, width=1920, bitrate=2808, streams=[backend]
        )
    ]
    dl = downloader({'a': create_clip(flavors)})

    res = dl.download_clips('', simple.io, simple.filters)

    assert res == RD_SUCCESS
    assert backend.save_stream.call_count == 3


def test_download_fails_if_too_many_failures(simple):
    backend = backend_that_fails_n_times(10)
    flavors = [
        StreamFlavor(
            media_type='video', height=1080, width=1920, bitrate=2808, streams=[backend]
        )
    ]
    dl = downloader({'a': create_clip(flavors)})

    res = dl.download_clips('', simple.io, simple.filters)

    assert res == RD_FAILED
    assert backend.save_stream.call_count >= 4


def test_pipe_does_not_retry(simple):
    backend = backend_that_fails_n_times(1)
    flavors = [
        StreamFlavor(
            media_type='video', height=1080, width=1920, bitrate=2808, streams=[backend]
        )
    ]
    dl = downloader({'a': create_clip(flavors)})

    res = dl.pipe('', simple.io, simple.filters)

    assert res == RD_FAILED



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from yledl import RD_SUCCESS
from yledl.backends import DASHHLSBackend, WgetBackend
from utils import MockIOContext


tv1_url = 'https://yletv-lh.akamaihd.net/i/yletv1hls_1@103188/master.m3u8'
io = MockIOContext(destdir='/tmp/')


class MockHLSBackend(DASHHLSBackend):
    def __init__(self, url, long_probe=False, program_id=None, is_live=False):
        super().__init__(url, long_probe, program_id, is_live)

        self.executed_commands = None

    def external_downloader(self, commands, env=None):
        self.executed_commands = commands
        return RD_SUCCESS

    def full_stream_already_downloaded(self, filename, clip, io):
        return False


class MockWgetBackend(WgetBackend):
    def __init__(self, url, file_extension):
        super().__init__(url, file_extension)

        self.executed_commands = None

    def external_downloader(self, commands, env=None):
        self.executed_commands = commands
        return RD_SUCCESS

    def full_stream_already_downloaded(self, filename, clip, io):
        return False


def test_hls_backend_save_stream():
    backend = MockHLSBackend(tv1_url, program_id=0)

    res = backend.save_stream('test.mkv', clip=None, io=io)

    assert res == RD_SUCCESS
    assert len(backend.executed_commands) == 1
    assert tv1_url in backend.executed_commands[0]


def test_hls_backend_pipe():
    backend = MockHLSBackend(tv1_url, program_id=0)

    res = backend.pipe(io)

    assert res == RD_SUCCESS
    assert len(backend.executed_commands) == 1
    assert tv1_url in backend.executed_commands[0]


def test_wget_backend_save_stream():
    backend = MockWgetBackend(tv1_url, file_extension='.mkv')

    res = backend.save_stream('test.mkv', clip=None, io=io)

    assert res == RD_SUCCESS
    assert len(backend.executed_commands) == 1
    assert tv1_url in backend.executed_commands[0]


def test_wget_backend_pipe():
    backend = MockWgetBackend(tv1_url, file_extension='.mkv')

    res = backend.pipe(io)

    assert res == RD_SUCCESS
    assert len(backend.executed_commands) == 1
    assert tv1_url in backend.executed_commands[0]



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import pytest
from datetime import datetime
from utils import fetch_title, fetch_stream_url, fetch_episode_pages, fetch_metadata
from yledl import StreamFilters


def test_areena_html5_stream_url():
    streamurl = fetch_stream_url('https://areena.yle.fi/1-787136')

    assert len(streamurl) == 1
    assert '.m3u8?' in streamurl[0]


def test_areena_html5_metadata():
    metadata = fetch_metadata('https://areena.yle.fi/1-787136')

    assert len(metadata) == 1
    flavors = metadata[0]['flavors']
    assert len(flavors) >= 4
    assert any(f.get('media_type') == 'video' for f in flavors)
    assert all('bitrate' in f for f in flavors)
    assert any('height' in f and 'width' in f for f in flavors)
    assert metadata[0]['duration_seconds'] == 907
    assert metadata[0]['region'] == 'World'
    assert metadata[0]['publish_timestamp'] == '2021-04-01T00:01:00+03:00'
    assert 'expired_timestamp' not in metadata[0]
    assert len(metadata[0]['description']) > 150
    assert metadata[0]['thumbnail'].startswith('https://images.cdn.yle.fi/')


def test_metadata_language():
    meta_fin = fetch_metadata('https://areena.yle.fi/1-403848', meta_language='fin')
    title_fin = meta_fin[0].get('title')
    assert title_fin.startswith('Suomen tie jatkosotaan')

    meta_swe = fetch_metadata('https://areena.yle.fi/1-403848', meta_language='swe')
    title_swe = meta_swe[0].get('title')
    assert title_swe.startswith('Finlands väg till fortsättningskriget')


def test_areena_series_titles():
    titles = fetch_title('https://areena.yle.fi/1-3826480')

    assert len(titles) == 10
    assert all(['Suomi on ruotsalainen' in t for t in titles])


def test_areena_series_urls():
    urls = fetch_stream_url('https://areena.yle.fi/1-3826480')

    assert len(urls) == 10
    assert all(['.m3u8' in url for url in urls])


@pytest.mark.geoblocked
def test_areena_live_url():
    streamurl = fetch_stream_url('tv1')

    assert len(streamurl) == 1
    assert '.m3u8' in streamurl[0]


@pytest.mark.geoblocked
def test_areena_live_metadata():
    metadata = fetch_metadata('tv1')

    assert len(metadata) == 1
    assert len(metadata[0]['flavors']) >= 1
    assert all(f.get('media_type') == 'video' for f in metadata[0]['flavors'])
    assert metadata[0]['region'] == 'Finland'


def test_areena_html5_clip_title():
    title = fetch_title('https://areena.yle.fi/1-3523087')

    assert len(title) == 1
    assert 'Metsien kätkemä' in title[0]


def test_areena_html5_clip_stream_url():
    streamurl = fetch_stream_url('https://areena.yle.fi/1-3523087')

    assert len(streamurl) == 1
    assert '.m3u8?' in streamurl[0]


# @pytest.mark.xfail(reason='This video has been broken in Areena since July 2022')
# def test_areena_awsmpodamdipv4_stream_url():
#     streamurl = fetch_stream_url('https://areena.yle.fi/1-50875269')
#
#     assert len(streamurl) == 1
#     assert '/index.m3u8?' in streamurl[0]
#
#
# @pytest.mark.xfail(reason='This video has been broken in Areena since July 2022')
# def test_areena_awsmpodamdipv4_metadata():
#     metadata = fetch_metadata('https://areena.yle.fi/1-50875269')
#
#     assert len(metadata) == 1
#     flavors = metadata[0]['flavors']
#     assert len(flavors) >= 1
#     assert all(f.get('media_type') == 'video' for f in flavors)
#     assert metadata[0]['duration_seconds'] == 257
#     assert metadata[0]['region'] == 'World'
#     assert metadata[0]['publish_timestamp'] == '2021-06-11T08:40:00+03:00'
#     assert 'expired_timestamp' not in metadata[0]
#     assert len(metadata[0]['description']) > 150


def test_areena_episode_pages():
    episodes = fetch_episode_pages('https://areena.yle.fi/1-3148871')

    # The first page contains 12 episodes, make sure we get several pages
    assert len(episodes) > 20


def test_areena_episode_pages_swedish():
    # Regression test for #336
    episodes = fetch_episode_pages('https://arenan.yle.fi/1-4583749')

    assert len(episodes) > 20


def test_areena_season_id():
    # An Areena TV URL can point to either a series or a season.
    #
    # For example, the series page of "Sekaisin" is 1-3430975 and the third
    # season of "Sekaisin" is 1-50570316.
    #
    # Yle-dl downloads all episodes from all seasons no matter which ID is
    # used.
    episodes = fetch_episode_pages('https://areena.yle.fi/1-50570316')

    # The seasons contain 8 to 15 episodes, make sure we get several seasons
    assert len(episodes) > 20


def test_areena_sort_by_season_episode():
    # These clips have season and episode numbers
    metadata = fetch_metadata('https://areena.yle.fi/1-4530023')

    # Should be sorted from oldest to newest
    timestamps = [x['publish_timestamp'] for x in metadata]
    assert len(timestamps) > 1
    assert timestamps == sorted(timestamps)


def test_areena_sort_by_timestamp():
    # Most of these clips have a timestamp (but not all do).
    # Most clips do not have an episode number (but some do).
    metadata = fetch_metadata('https://areena.yle.fi/1-3830094')

    # Should be sorted from oldest to newest
    timestamps = [x.get('publish_timestamp', '') for x in metadata]

    assert len(timestamps) > 1
    assert timestamps == sorted(timestamps)


def test_areena_season_and_episode_number():
    # The episode titles should include S01E01, S01E02, etc.
    # These episodes have season number in their description, e.g. "Kausi 1"
    expected_episodes = []
    for season in range(1, 4 + 1):
        number_of_episodes = 12 if season < 4 else 10
        for episode in range(1, number_of_episodes + 1):
            expected_episodes.append(f'S{season:02d}E{episode:02d}')

    titles = fetch_title('https://areena.yle.fi/1-4530023')

    assert len(titles) == len(expected_episodes)
    for title, substring in zip(titles, expected_episodes):
        assert substring in title


def test_areena_season_number_in_webpage():
    # These episode don't have the season number in their description but do
    # have it as part of the web page HTML
    expected_episodes = []
    season = 1
    for episode in range(1, 7 + 1):
        expected_episodes.append(f'S{season:02d}E{episode:02d}')

    titles = fetch_title('https://areena.yle.fi/1-50831169')

    assert len(titles) == len(expected_episodes)
    for title, substring in zip(titles, expected_episodes):
        assert substring in title


def test_areena_latest_episode():
    # This series has seasons
    filters = StreamFilters(latest_only=True)
    metadata = fetch_metadata('https://areena.yle.fi/1-4655342', filters)

    assert len(metadata) == 1

    # The latest episode at the time of writing this test was
    # published on 2021-01-27
    publish_date = datetime.strptime(metadata[0]['publish_timestamp'][:10], '%Y-%m-%d')
    assert publish_date >= datetime(2021, 1, 27)


def test_areena_latest_episode_no_seasons():
    # This series has no seasons
    filters = StreamFilters(latest_only=True)
    metadata = fetch_metadata('https://areena.yle.fi/1-1494772', filters)

    assert len(metadata) == 1

    # The latest episode at the time of writing this test was
    # published on 2022-07-12
    publish_date = datetime.strptime(metadata[0]['publish_timestamp'][:10], '%Y-%m-%d')
    assert publish_date >= datetime(2022, 7, 12)


def test_areena_package_page():
    # This is a "package" type page
    episodes = fetch_episode_pages('https://areena.yle.fi/tv/ohjelmat/30-1774')

    assert len(episodes) > 10


def test_areena_uusimmat_tab_page():
    # This is a "TVView" page, for example MOT or Perjantai.
    # The page has multiple tabs listing the same episodes in different ways.
    # Should download the episodes from the "Uusimmat" tab.
    episodes = fetch_episode_pages('https://areena.yle.fi/tv/ohjelmat/57-4Nj71REaB')

    assert len(episodes) > 10


def test_areena_clips_only():
    # This page has only the "Klipit" tab, not "Jaksot" like most pages
    episodes = fetch_episode_pages('https://areena.yle.fi/1-4354262')

    assert len(episodes) >= 5


def test_areena_clip_tab_selected():
    titles = fetch_title('https://areena.yle.fi/1-4530023?t=klipit')

    assert len(titles) >= 6
    assert all([t.startswith('Modernit miehet') for t in titles])



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from typing import List, Optional
from datetime import datetime

from dataclasses import dataclass, field
from .io import OutputFileNameGenerator

from .areena_playlist_parser import AreenaPlaylistParser
from .streamflavor import failed_flavor

## Clip


@dataclass
class Clip:
    webpage: str
    flavors: list = field(default_factory=list)
    title: str = ''
    episode_title: str = ''
    description: Optional[str] = None
    duration_seconds: Optional[int] = None
    region: str = 'Finland'
    publish_timestamp: Optional[datetime] = None
    expiration_timestamp: Optional[datetime] = None
    subtitles: List = field(default_factory=list)
    program_id: Optional[str] = None
    origin_url: Optional[str] = None
    thumbnail: Optional[str] = None

    def metadata(self, io):
        flavors_meta = sorted(
            (self.flavor_meta(f) for f in self.flavors),
            key=lambda x: x.get('bitrate', 0),
        )
        meta = [
            ('program_id', self.program_id),
            ('webpage', self.webpage),
            ('title', self.title),
            ('episode_title', self.episode_title),
            ('description', self.description),
            ('filename', self.meta_file_name(self.flavors, io)),
            ('flavors', flavors_meta),
            ('duration_seconds', self.duration_seconds),
            (
                'subtitles',
                [
                    {'language': x.lang, 'url': x.url, 'category': x.category}
                    for x in self.subtitles
                ],
            ),
            ('thumbnail', self.thumbnail),
            ('region', self.region),
            ('publish_timestamp', self.format_timestamp(self.publish_timestamp)),
            ('expiration_timestamp', self.format_timestamp(self.expiration_timestamp)),
        ]
        return self.ignore_none_values(meta)

    def meta_file_name(self, flavors, io):
        flavors = sorted(flavors, key=lambda x: x.bitrate or 0)
        flavors = [fl for fl in flavors if any(s.is_valid() for s in fl.streams)]
        if flavors:
            extensions = [
                s.file_extension('mkv') for s in flavors[-1].streams if s.is_valid()
            ]
            if extensions:
                return OutputFileNameGenerator().filename(self.title, extensions[0], io)

        return None

    def format_timestamp(self, ts):
        return ts.isoformat() if ts else None

    def flavor_meta(self, flavor):
        if all(not s.is_valid() for s in flavor.streams):
            return self.error_flavor_meta(flavor)
        else:
            return self.valid_flavor_meta(flavor)

    def valid_flavor_meta(self, flavor):
        backends = [s.name for s in flavor.streams if s.is_valid()]

        streams = flavor.streams
        if streams and any(s.is_valid() for s in streams):
            valid_stream = next(s for s in streams if s.is_valid())
            url = valid_stream.stream_url()
        else:
            url = None

        meta = [
            ('media_type', flavor.media_type),
            ('height', flavor.height),
            ('width', flavor.width),
            ('bitrate', flavor.bitrate),
            ('backends', backends),
            ('url', url),
        ]
        return self.ignore_none_values(meta)

    def error_flavor_meta(self, flavor):
        error_messages = [
            s.error_message
            for s in flavor.streams
            if not s.is_valid() and s.error_message
        ]
        if error_messages:
            msg = error_messages[0]
        else:
            msg = 'Unknown error'

        return {'error': msg}

    def ignore_none_values(self, li):
        return {key: value for (key, value) in li if value is not None}


class FailedClip(Clip):
    def __init__(self, webpage, error_message, **kwargs):
        super().__init__(
            webpage=webpage, flavors=[failed_flavor(error_message)], **kwargs
        )


class ClipExtractor:
    def __init__(self, httpclient):
        self.httpclient = httpclient

    def extract(self, url, latest_only):
        playlist = self.get_playlist(url, latest_only)
        return (self.extract_clip(clipurl, url) for clipurl in playlist)

    def get_playlist(self, url, latest_only=False):
        return AreenaPlaylistParser(self.httpclient).get(url, latest_only)

    def extract_clip(self, url, origin_url):
        raise NotImplementedError('extract_clip must be overridden')



# This file is part of yle-dl.
#
# Copyright 2010-2024 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
import json
import re
from typing import List

import logging
from .areena_api import EpisodeMetadata
from .http import update_url_query
from .play_list_data import PlaylistData
from .timestamp import parse_areena_timestamp

logger = logging.getLogger('yledl')


class AreenaPlaylistParser:
    """Get a list of episodes in a series from Areena API

    Reference: https://docs.api.yle.fi/api/programs-api-v3
    """

    def __init__(self, httpclient):
        self.httpclient = httpclient

    def get(self, url, latest_only=False):
        """If url is a series page, return a list of included episode pages."""
        tree = self.httpclient.download_html_tree(url)
        if tree is None:
            logger.warning(f'Failed to download {url} while looking for a playlist')
            return [url]

        playlist = []
        playlist_data = None
        if self._is_tv_series_page(tree):
            logger.debug('TV playlist')
            playlist_data = self._parse_series_playlist(tree)
        elif self._is_radio_series_page(tree):
            logger.debug('Radio playlist')
            playlist_data = self._parse_radio_playlist(tree)
        elif self._extract_package_id(tree) is not None:
            logger.debug('Package playlist')
            playlist_data = self._parse_package_playlist(tree)
        else:
            logger.debug('Not a playlist')
            playlist = [url]

        if playlist_data is not None:
            playlist = self._download_playlist_or_latest(playlist_data, latest_only)
            logger.debug(f'playlist page with {len(playlist)} episodes')

        return playlist

    def _is_tv_series_page(self, tree):
        next_data_tag = tree.xpath('//script[@id="__NEXT_DATA__"]')
        if len(next_data_tag) == 0:
            return False

        next_data = json.loads(next_data_tag[0].text)
        ptype = (
            next_data.get('props', {})
            .get('pageProps', {})
            .get('meta', {})
            .get('item', {})
            .get('type')
        )

        return ptype in ['TVSeries', 'TVSeason', 'TVView', 'RadioSeries', 'Package']

    def _is_radio_series_page(self, tree):
        is_radio_page = len(tree.xpath('//div[contains(@class, "RadioPlayer")]')) > 0
        if is_radio_page:
            episode_modal = tree.xpath('//div[starts-with(@class, "EpisodeModal")]')
            play_button = tree.xpath(
                '//main//button[starts-with(@class, "PlayButton")]'
            )
            return not episode_modal and not play_button
        else:
            return False

    def _parse_series_playlist(self, html_tree):
        next_data_tag = html_tree.xpath('//script[@id="__NEXT_DATA__"]')
        if next_data_tag:
            next_data = json.loads(next_data_tag[0].text)
            page_props = next_data.get('props', {}).get('pageProps', {})
            tabs = page_props.get('view', {}).get('tabs', [])
            first_tab_slug = tabs[0].get('slug') if tabs else None
            selected_tab = page_props.get('selectedTab') or first_tab_slug or 'jaksot'
            return self._parse_episodes_tab(
                tabs, selected_tab
            ) or self._parse_episodes_tab(tabs, None)

        return None

    def _parse_episodes_tab(self, next_data_tabs, tab_slug):
        if tab_slug:
            episodes_tab = [
                tab for tab in next_data_tabs if tab.get('slug') == tab_slug
            ]
        else:
            episodes_tab = [
                tab
                for tab in next_data_tabs
                if tab.get('type') == 'tab' and 'title' not in tab
            ]

        if episodes_tab:
            episodes_content = episodes_tab[0].get('content', [])
            if episodes_content:
                playlist_data = episodes_content[0]
                if playlist_data.get('title') not in ['Katso myös', 'Kuuntele myös']:
                    uri = playlist_data.get('source', {}).get('uri')

                    series_parameters = {}
                    filters = playlist_data.get('filters', [])
                    if filters:
                        options = filters[0].get('options', [])
                        series_parameters = [x['parameters'] for x in options]

                    return PlaylistData(uri, series_parameters)

        return None

    def _parse_package_playlist(self, html_tree):
        package_tag = html_tree.xpath('//div[@class="package-view"]/@data-view')
        if package_tag:
            package_data = json.loads(package_tag[0])
            tabs = package_data.get('tabs', [])
            if tabs:
                content = tabs[0].get('content', [])
                if content:
                    uri = content[0].get('source', {}).get('uri')
                    return PlaylistData(uri, [])

        return None

    def _parse_radio_playlist(self, html_tree):
        state_tag = html_tree.xpath(
            '//script[contains(., "window.STORE_STATE_FROM_SERVER")]'
        )
        if state_tag:
            state_str = state_tag[0].text
            data = json.loads(state_str.split('=', 1)[-1].strip())
            tabs = data.get('viewStore', {}).get('viewPageView', {}).get('tabs', [])
            tabs = [t for t in tabs if t.get('title') in ['Jaksot', 'Avsnitt']]
            if tabs:
                all_content = tabs[0].get('allContent')
                if all_content:
                    uri = all_content[0].get('source', {}).get('uri')
                    return PlaylistData(uri, [])

        return None

    def _download_playlist_or_latest(self, playlist_data, latest_only):
        season_urls = list(enumerate(playlist_data.season_playlist_urls(), start=1))
        if latest_only:
            # Optimization: The latest episode belongs to the latest season
            season_urls = season_urls[-1:]

        playlist = self._download_playlist(season_urls)

        # Heuristics: If most episodes do not have an episode number,
        # use time-based sorting.
        if self._episode_numbers_are_rare(playlist) and self._timestamps_are_common(
            playlist
        ):
            playlist = [x.with_episode_number(None) for x in playlist]

        # We can't control whether Areena API returns episodes in
        # ascending or descending order. Additionally, metadata
        # contains only the date (not hours or minutes) so it's not
        # possible to sort intra-day episodes properly. This is a hack
        # that tries to sort intra-day episodes in ascending order.
        # For example: https://areena.yle.fi/1-3863205
        if self._is_descending_date_based_playlist(playlist):
            playlist = reversed(playlist)

        # Sort in ascending order: first by episode number, then by date
        playlist = sorted(playlist, key=lambda x: x.sort_key())

        # The episode API doesn't seem to have any way to download only the
        # latest episode or start from the latest. We need to download all and
        # pick the latest.
        if latest_only:
            playlist = playlist[-1:]

        return [x.uri for x in playlist]

    def _episode_numbers_are_rare(self, playlist):
        num_has_episode = sum(p.episode_number is not None for p in playlist)
        return num_has_episode < 0.5 * len(playlist)

    def _timestamps_are_common(self, playlist):
        num_has_timestamp = sum(p.release_date is not None for p in playlist)
        return num_has_timestamp > 0.8 * len(playlist)

    def _is_descending_date_based_playlist(self, playlist):
        if not all(p.episode_number is None for p in playlist):
            return False

        prev_ts = None
        for p in playlist:
            if (
                prev_ts is not None
                and p.release_date is not None
                and p.release_date < prev_ts
            ):
                return True

            prev_ts = p.release_date

        return False

    def _download_playlist(self, season_urls):
        playlist = []
        for season_num, season_url in season_urls:
            # Areena server fails (502 Bad gateway) if page_size is larger
            # than 100.
            page_size = 100
            offset = 0
            has_next_page = True
            while has_next_page:
                logger.debug(
                    f'Getting a playlist page, season = {season_num}, '
                    f'size = {page_size}, offset = {offset}'
                )

                params = {
                    'offset': str(offset),
                    'limit': str(page_size),
                    'app_id': 'areena-web-items',
                    'app_key': 'wlTs5D9OjIdeS9krPzRQR4I1PYVzoazN',
                }
                playlist_page_url = update_url_query(season_url, params)
                page = self._parse_series_episode_data(playlist_page_url, season_num)

                if page is None:
                    logger.warning(
                        f'Playlist failed at offset {offset}. Some episodes may be missing!'
                    )
                    break

                playlist.extend(page)
                offset += len(page)
                has_next_page = len(page) == page_size

        return playlist

    def _parse_series_episode_data(self, playlist_page_url, season_number):
        playlist = self.httpclient.download_json(playlist_page_url)
        if playlist is None:
            return None

        episodes = []
        for data in playlist.get('data', []):
            uri = self._episode_uri(data)
            episode_number = self._episode_number(data)
            release_date = self._tv_release_date(data) or self._radio_release_date(data)

            if uri:
                episodes.append(
                    EpisodeMetadata(uri, season_number, episode_number, release_date)
                )

        return episodes

    @staticmethod
    def _extract_package_id(tree):
        package_id = tree.xpath('/html/body/@data-package-id')
        if package_id:
            return package_id[0]
        else:
            return None

    @staticmethod
    def _episode_uri(data):
        program_uri = data.get('pointer', {}).get('uri')
        if program_uri:
            media_id = program_uri.rsplit('/')[-1]
            return f'https://areena.yle.fi/{media_id}'
        else:
            return None

    @staticmethod
    def _episode_number(data):
        title = data.get('title')
        if title:
            # Try to parse the episode number from the title. That's the
            # only location where the episode number is available in the
            # API response.
            m = re.match(r'Jakso (\d+)', title, flags=re.IGNORECASE)
            if m:
                return int(m.group(1))

        return None

    def _tv_release_date(self, data):
        labels = data.get('labels')
        generics = self._label_by_type(labels, 'generic', 'formatted')
        for val in generics:
            # Look for a label that matches the format "pe 15.3.2019"
            m = re.match(
                r'[a-z]{2} (?P<day>\d{1,2})\.(?P<month>\d{1,2})\.(?P<year>\d{4})', val
            )
            if m:
                return datetime(
                    int(m.group('year')), int(m.group('month')), int(m.group('day'))
                )

        return None

    def _radio_release_date(self, data):
        labels = data.get('labels')
        date_str = self._label_by_type(labels, 'releaseDate', 'raw')
        if date_str:
            try:
                return parse_areena_timestamp(date_str[0])
            except ValueError:
                pass

        return None

    def _label_by_type(self, labels: dict, type_name: str, key_name: str) -> List[str]:
        """Return a key value of an Areena API label object which as the given type."""
        matches = [x for x in labels if x.get('type') == type_name]
        return [x[key_name] for x in matches if key_name in x]



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime
import logging
import re
from .localization import TranslationChooser
from .subtitles import Subtitle
from .timestamp import parse_areena_timestamp, format_finnish_short_weekday_and_date


logger = logging.getLogger('yledl')


class AreenaPreviewApiParser:
    def __init__(self, data):
        self.preview = data or {}

    def media_id(self):
        ongoing = self.ongoing()
        mid1 = ongoing.get('media_id')
        mid2 = ongoing.get('adobe', {}).get('yle_media_id')
        return mid1 or mid2

    def duration_seconds(self):
        return self.ongoing().get('duration', {}).get('duration_in_seconds')

    def title(self, language_chooser):
        title = {}
        ongoing = self.ongoing()
        title_object = ongoing.get('title', {})
        if title_object:
            title['title'] = language_chooser.choose_long_form(title_object).strip()

        series_title_object = ongoing.get('series', {}).get('title', {})
        if series_title_object:
            title['series_title'] = language_chooser.choose_long_form(
                series_title_object
            ).strip()

        # If title['title'] does not equal title['episode_title'], then
        # the episode title is title['title'].
        #
        # If title['title'] equals title['episode_title'], then either
        # 1. the episode title is the publication date ("pe 16.9.2022"), or
        # 2. the episode title is title['title']
        #
        # It seem impossible to decide which of the cases 1. or 2. should apply
        # based on the preview API response only. We will always use the date
        # (case 1.) because that is the more common case.
        if title.get('title') is not None and title.get('title') == title.get(
            'series_title'
        ):
            title_timestamp = parse_areena_timestamp(ongoing.get('start_time'))
            if title_timestamp:
                # Should be localized (Finnish or Swedish) based on language_chooser
                title['title'] = format_finnish_short_weekday_and_date(title_timestamp)

        return title

    def description(self, language_chooser):
        description_object = self.ongoing().get('description', {})
        if not description_object:
            return None

        description_text = language_chooser.choose_long_form(description_object) or ''
        return description_text.strip()

    def season_and_episode(self):
        res = {}
        episode = self.ongoing().get('episode_number')
        if episode is not None:
            res = {'episode': episode}

            desc = self.description(TranslationChooser(['fin'])) or ''
            m = re.match(r'Kausi (\d+)\b', desc)
            if m:
                res.update({'season': int(m.group(1))})

        return res

    def available_at_region(self):
        return self.ongoing().get('region')

    def timestamp(self):
        if self.is_live():
            return datetime.now().replace(microsecond=0)
        else:
            dt = self.ongoing().get('start_time')
            return parse_areena_timestamp(dt)

    def manifest_url(self):
        return self.ongoing().get('manifest_url')

    def media_url(self):
        return self.ongoing().get('media_url')

    def media_type(self):
        if not self.preview:
            return None
        elif self.ongoing().get('content_type') == 'AudioObject':
            return 'audio'
        else:
            return 'video'

    def is_live(self):
        data = self.preview.get('data', {})
        return 'ongoing_channel' in data or 'ongoing_event' in data

    def is_pending(self):
        data = self.preview.get('data', {})
        pending = data.get('pending_event') or data.get('pending_ondemand')
        return pending is not None

    def is_expired(self):
        data = self.preview.get('data', {})
        return data.get('gone') is not None

    def ongoing(self):
        data = self.preview.get('data', {})
        return (
            data.get('ongoing_ondemand')
            or data.get('ongoing_event', {})
            or data.get('ongoing_channel', {})
            or data.get('pending_event')
            or {}
        )

    def subtitles(self):
        langname2to3 = {
            'fi': 'fin',
            'fih': 'fin',
            'sv': 'swe',
            'svh': 'swe',
            'se': 'smi',
            'en': 'eng',
        }
        hearing_impaired_langs = ['fih', 'svh']

        sobj = self.ongoing().get('subtitles', [])
        subtitles = []
        for s in sobj:
            # Areena has two subtitle objects. The newer object has "language"
            # and "kind" properties. "language" is a three-letter language code.
            lcode_longform = s.get('language', None)
            # The older (not used anymore as of Nov 2023?) format has "lang",
            # which is a two-letter language code with a possible third letter
            # "h" indicating hard-of-hearing subtitles.
            lcode = s.get('lang', None)

            if lcode_longform:
                lang = lcode_longform
                if s.get('kind', None) == 'hardOfHearing':
                    category = 'ohjelmatekstitys'
                else:
                    category = 'käännöstekstitys'
            elif lcode:
                lang = langname2to3.get(lcode, lcode)
                if lcode in hearing_impaired_langs:
                    category = 'ohjelmatekstitys'
                else:
                    category = 'käännöstekstitys'
            else:
                lang = 'unk'
                category = 'käännöstekstitys'
            url = s.get('uri', None)
            if lang and url:
                subtitles.append(Subtitle(url, lang, category))
        return subtitles

    def thumbnail_url(self):
        image_data = self.ongoing().get('image', {})
        if 'id' not in image_data:
            return None

        version = image_data['version'] or 1624522786
        image_id = image_data['id']

        return f'https://images.cdn.yle.fi/image/upload/f_auto,c_limit,w_1080,q_auto/v{version}/{image_id}'



# This file is part of yle-dl.
#
# Copyright 2010-2022 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

import pytest
from datetime import datetime, date
from utils import FixedOffset
from yledl.titleformatter import TitleFormatter

tf = TitleFormatter()


@pytest.fixture
def pasila():
    return {
        'title': 'Vanhempainyhdistys',
        'publish_timestamp': datetime(2018, 4, 12, 16, 30, 45, tzinfo=FixedOffset(2)),
        'series_title': 'Pasila',
        'subheading': 'tekstitys englanniksi',
        'program_id': '1-86743',
        'season': 1,
        'episode': 3,
    }


def test_none_title():
    assert tf.format(None) is None


def test_title_timestamp():
    title = tf.format('test', publish_timestamp=datetime(2018, 1, 2, 3, 4, 5))
    assert title == 'test: 2018-01-02T03:04'


def test_title_date_only():
    title = tf.format('test', publish_timestamp=date(2018, 1, 2))
    assert title == 'test: 2018-01-02'


def test_repeated_main_title():
    title = tf.format(
        'Uutiset: Uutiset iltapäivällä', publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'Uutiset iltapäivällä: 2021-01-01'


def test_subheading():
    title = tf.format(
        'EM-kisat', subheading='Kymmenottelu', publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'EM-kisat: Kymmenottelu: 2021-01-01'


def test_no_repeated_subheading():
    title = tf.format(
        'Uutiset: Kymmenen uutiset',
        subheading='Uutiset',
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Uutiset: Kymmenen uutiset: 2021-01-01'


def test_season_and_episode():
    title = tf.format(
        'Isänmaan toivot', season=2, episode=6, publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'Isänmaan toivot: S02E06-2021-01-01'


def test_episode_without_season():
    title = tf.format('Isänmaan toivot', episode=12, publish_timestamp=date(2021, 1, 1))
    assert title == 'Isänmaan toivot: E12-2021-01-01'


def test_remove_genre_prefix():
    assert (
        tf.format('Elokuva: Indiana Jones', publish_timestamp=date(2021, 1, 1))
        == 'Indiana Jones: 2021-01-01'
    )


def test_series_title():
    title = tf.format(
        'Kerblam!', series_title='Doctor Who', publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'Doctor Who: Kerblam!: 2021-01-01'


def test_no_repeated_series_title():
    title = tf.format(
        'Doctor Who', series_title='Doctor Who', publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'Doctor Who: 2021-01-01'


def test_no_repeated_series_title_whole_words():
    title = tf.format(
        'Noin viikon studion uusi vuosi',
        series_title='Noin viikon studio',
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Noin viikon studio: Noin viikon studion uusi vuosi: 2021-01-01'


def test_no_repeated_series_title_with_subheading():
    title = tf.format(
        'Solsidan',
        series_title='Solsidan',
        subheading='Nya avsnitt från Solsidan',
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Solsidan: Nya avsnitt från Solsidan: 2021-01-01'


def test_no_repeated_series_title_with_episode_title():
    title = tf.format(
        'Doctor Who: Kerblam!',
        series_title='Doctor Who',
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Doctor Who: Kerblam!: 2021-01-01'


def test_series_name_as_part_of_episode_title():
    title = tf.format(
        'Rölli ja Robotti Ruttunen',
        series_title='Rölli',
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Rölli: Rölli ja Robotti Ruttunen: 2021-01-01'


def test_main_title_equals_series_title_plus_age_limit():
    title = tf.format(
        'Rantahotelli (S)',
        series_title='Rantahotelli',
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Rantahotelli: 2021-01-01'


def test_strip_whitespace():
    title = tf.format(' Rantahotelli ', publish_timestamp=date(2021, 1, 1))
    assert title == 'Rantahotelli: 2021-01-01'

    title = tf.format(
        'Uutiset klo 18', series_title='Uutiset ', publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'Uutiset: Uutiset klo 18: 2021-01-01'

    title = tf.format(
        'Uutiset klo 18', series_title=' Uutiset ', publish_timestamp=date(2021, 1, 1)
    )
    assert title == 'Uutiset: Uutiset klo 18: 2021-01-01'


def test_all_components(pasila):
    assert (
        tf.format(**pasila) == 'Pasila: Vanhempainyhdistys: '
        'tekstitys englanniksi: S01E03-2018-04-12T16:30'
    )


def test_template(pasila):
    fmt = TitleFormatter('${series}: ${title}: ${episode}-${timestamp}')
    assert (
        fmt.format(**pasila) == 'Pasila: Vanhempainyhdistys: '
        'tekstitys englanniksi: S01E03-2018-04-12T16:30'
    )

    fmt = TitleFormatter('${series}: ${episode}-${timestamp}')
    assert fmt.format(**pasila) == 'Pasila: S01E03-2018-04-12T16:30'

    fmt = TitleFormatter('${title}-${timestamp}')
    assert (
        fmt.format(**pasila) == 'Vanhempainyhdistys: '
        'tekstitys englanniksi-2018-04-12T16:30'
    )

    fmt = TitleFormatter('${timestamp}: ${title}')
    assert (
        fmt.format(**pasila) == '2018-04-12T16:30: '
        'Vanhempainyhdistys: tekstitys englanniksi'
    )

    fmt = TitleFormatter('${program_id}: ${series}')
    assert fmt.format(**pasila) == '1-86743: Pasila'

    fmt = TitleFormatter('${series}-${program_id}')
    assert fmt.format(**pasila) == 'Pasila-1-86743'


def test_template_date(pasila):
    fmt = TitleFormatter('${series}-${date}')
    assert fmt.format(**pasila) == 'Pasila-2018-04-12'

    fmt = TitleFormatter('${date}: ${series}')
    assert fmt.format(**pasila) == '2018-04-12: Pasila'

    fmt = TitleFormatter('${series}-${date}-${timestamp}')
    assert fmt.format(**pasila) == 'Pasila-2018-04-12-2018-04-12T16:30'


def test_template_series_duplicated_in_main_title():
    data = {
        'title': 'Jopet-show: Haikeaa joulua',
        'series_title': 'Jopet-show',
        'publish_timestamp': datetime(2018, 4, 12, 16, 30, 45, tzinfo=FixedOffset(2)),
    }

    fmt = TitleFormatter('${series}')
    assert fmt.format(**data) == 'Jopet-show'

    fmt = TitleFormatter('${title}')
    assert fmt.format(**data) == 'Haikeaa joulua'

    fmt = TitleFormatter('${series}: ${title}')
    assert fmt.format(**data) == 'Jopet-show: Haikeaa joulua'


def test_template_series_equals_main_title():
    data = {
        'title': 'Rantahotelli',
        'series_title': 'Rantahotelli',
        'publish_timestamp': datetime(2018, 4, 12, 16, 30, 45, tzinfo=FixedOffset(2)),
    }

    fmt = TitleFormatter('${series}')
    assert fmt.format(**data) == ''

    fmt = TitleFormatter('${title}')
    assert fmt.format(**data) == 'Rantahotelli'

    fmt = TitleFormatter('${series}${title}')
    assert fmt.format(**data) == 'Rantahotelli'


def test_template_literal(pasila):
    fmt = TitleFormatter('Areena ${series}-${episode}')
    assert fmt.format(**pasila) == 'Areena Pasila-S01E03'

    fmt = TitleFormatter('${series}: Areena-${episode}')
    assert fmt.format(**pasila) == 'Pasila: Areena-S01E03'

    fmt = TitleFormatter('${series}: ${episode} Areena')
    assert fmt.format(**pasila) == 'Pasila: S01E03 Areena'

    fmt = TitleFormatter('Areena: ${series} episode ${episode} Areena')
    assert fmt.format(**pasila) == 'Areena: Pasila episode S01E03 Areena'


def test_template_duplicate_key(pasila):
    fmt = TitleFormatter('${series}: ${series}')
    assert fmt.format(**pasila) == 'Pasila: Pasila'


def test_template_literal_dollar_sign(pasila):
    fmt = TitleFormatter('${series} 500$$ ABC')
    assert fmt.format(**pasila) == 'Pasila 500$ ABC'

    fmt = TitleFormatter('${series} 500$$ $$$$ ABC')
    assert fmt.format(**pasila) == 'Pasila 500$ $$ ABC'


def test_template_unbalanced_dollar_literals(pasila):
    fmt = TitleFormatter('${series} $$$ ABC')
    assert fmt.format(**pasila) == 'Pasila $$ ABC'


def test_unknown_templates_are_not_substituted(pasila):
    fmt = TitleFormatter('${series}: ${invalid}: ${timestamp}')
    assert fmt.format(**pasila) == 'Pasila: ${invalid}: 2018-04-12T16:30'


def test_unclosed_template(pasila):
    fmt = TitleFormatter('${series}: ${timestamp')
    assert fmt.format(**pasila) == 'Pasila: ${timestamp'

    fmt = TitleFormatter('${series}: ${title: ${timestamp}')
    assert fmt.format(**pasila) == 'Pasila: ${title: 2018-04-12T16:30'


def test_template_incorrectly_balanced_brackets(pasila):
    fmt = TitleFormatter('${series}: ${title: ${timestamp}}')
    assert fmt.format(**pasila) == 'Pasila: ${title: 2018-04-12T16:30}'


def test_title_and_date_placeholder():
    fmt = TitleFormatter(placeholder='NA')
    title = fmt.format('test', publish_timestamp=datetime(2018, 1, 2, 3, 4, 5))
    assert title == 'NA: test: NA-2018-01-02T03:04'


def test_series_title_placeholder():
    fmt = TitleFormatter(placeholder='NA')
    title = fmt.format(
        title='pe 1.1.2021',
        series_title='Isänmaan toivot',
        season=2,
        episode=6,
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Isänmaan toivot: pe 1.1.2021: S02E06-2021-01-01'


def test_title_only_placeholder():
    fmt = TitleFormatter(placeholder='NA')
    title = fmt.format('Kerblam!', series_title='Doctor Who', season=11, episode=7)
    assert title == 'Doctor Who: Kerblam!: S11E07-NA'


def test_template_variables_placeholder():
    fmt = TitleFormatter('${series}-${program_id}', placeholder='NA')
    title = fmt.format('Kerblam!', series_title='Doctor Who')
    assert title == 'Doctor Who-NA'

    fmt = TitleFormatter('${series}-${date}', placeholder='NA')
    title = fmt.format('Kerblam!', series_title='Doctor Who')
    assert title == 'Doctor Who-NA'

    fmt = TitleFormatter('${title}-${episode}-${date}', placeholder='NA')
    title = fmt.format('Kerblam!', series_title='Doctor Who')
    assert title == 'Kerblam!-NA-NA'


def test_no_repeated_series_title_placeholder():
    title = tf.format(
        'Solsidan',
        series_title='Solsidan',
        season=6,
        episode=3,
        publish_timestamp=date(2021, 1, 1),
    )
    assert title == 'Solsidan: S06E03-2021-01-01'



# This file is part of yle-dl.
#
# Copyright 2010-2024 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from yledl import YleDlDownloader, StreamFilters
from yledl.backends import BaseDownloader, Backends, FailingBackend
from yledl.streamflavor import StreamFlavor, failed_flavor
from yledl.titleformatter import TitleFormatter


class MockBackend(BaseDownloader):
    def __init__(self, name, data=None):
        super().__init__()
        self.name = name
        self.data = data

    def is_valid(self):
        return True


class MockGeoLocation:
    def located_in_finland(self, referrer):
        return True


flavors = [
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 1)],
        bitrate=190,
        width=224,
        height=126,
        media_type='',
    ),
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 5)],
        bitrate=1506,
        width=1280,
        height=720,
        media_type='',
    ),
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 6)],
        bitrate=2628,
        width=1280,
        height=720,
        media_type='',
    ),
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 7)],
        bitrate=4128,
        width=1920,
        height=1080,
        media_type='',
    ),
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 4)],
        bitrate=964,
        width=640,
        height=360,
        media_type='',
    ),
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 3)],
        bitrate=668,
        width=640,
        height=360,
        media_type='',
    ),
    StreamFlavor(
        streams=[MockBackend('ffmpeg', 2)],
        bitrate=469,
        width=640,
        height=360,
        media_type='',
    ),
]


def yle_dl_downloader():
    return YleDlDownloader(MockGeoLocation(), TitleFormatter(), None)


def video_flavor(streams):
    return StreamFlavor(media_type='video', streams=streams)


def filter_flavors(flavors, max_height=None, max_bitrate=None, enabled_backends=None):
    if enabled_backends is None:
        enabled_backends = list(Backends.default_order)

    filters = StreamFilters(
        maxheight=max_height, maxbitrate=max_bitrate, enabled_backends=enabled_backends
    )
    return yle_dl_downloader().select_flavor(flavors, filters)


def backend_data(flavor):
    return [s.data for s in flavor.streams]


def test_empty_input():
    assert filter_flavors([]) is None
    assert filter_flavors([], max_height=720) is None
    assert filter_flavors([], max_bitrate=5000) is None
    assert filter_flavors([], max_height=720, max_bitrate=5000) is None


def test_only_failed_flavors():
    failed_flavors = [
        failed_flavor('First failure'),
        failed_flavor('Second failure'),
        failed_flavor('Third failure'),
    ]

    res1 = filter_flavors(failed_flavors)
    assert res1 is not None
    assert len(res1.streams) == 1
    assert isinstance(res1.streams[0], FailingBackend)

    res2 = filter_flavors(failed_flavors, max_height=720)
    assert res2 is not None
    assert len(res2.streams) == 1
    assert isinstance(res2.streams[0], FailingBackend)

    res3 = filter_flavors(failed_flavors, max_bitrate=2000)
    assert res3 is not None
    assert len(res3.streams) == 1
    assert isinstance(res3.streams[0], FailingBackend)


def test_no_filters():
    assert backend_data(filter_flavors(flavors)) == [7]


def test_bitrate_filter():
    assert backend_data(filter_flavors(flavors, None, 10)) == [1]
    assert backend_data(filter_flavors(flavors, None, 200)) == [1]
    assert backend_data(filter_flavors(flavors, None, 963)) == [3]
    assert backend_data(filter_flavors(flavors, None, 964)) == [4]
    assert backend_data(filter_flavors(flavors, None, 1000)) == [4]
    assert backend_data(filter_flavors(flavors, None, 5000)) == [7]


def test_resolution_filter():
    assert backend_data(filter_flavors(flavors, 100, None)) == [1]
    assert backend_data(filter_flavors(flavors, 480, None)) == [2]
    assert backend_data(filter_flavors(flavors, 719, None)) == [2]
    assert backend_data(filter_flavors(flavors, 720, None)) == [5]
    assert backend_data(filter_flavors(flavors, 1080, None)) == [7]
    assert backend_data(filter_flavors(flavors, 2160, None)) == [7]


def test_combined_filters():
    assert backend_data(filter_flavors(flavors, 10, 10)) == [1]
    assert backend_data(filter_flavors(flavors, 200, 10)) == [1]
    assert backend_data(filter_flavors(flavors, 10, 200)) == [1]
    assert backend_data(filter_flavors(flavors, 360, 400)) == [1]
    assert backend_data(filter_flavors(flavors, 360, 650)) == [2]
    assert backend_data(filter_flavors(flavors, 360, 700)) == [3]
    assert backend_data(filter_flavors(flavors, 360, 5000)) == [4]
    assert backend_data(filter_flavors(flavors, 720, 200)) == [1]
    assert backend_data(filter_flavors(flavors, 2160, 1506)) == [5]


def test_combined_filter_with_some_failed_flavors():
    test_flavors = [
        failed_flavor('Failure'),
        StreamFlavor(
            streams=[MockBackend('ffmpeg', 2)],
            bitrate=190,
            width=224,
            height=126,
            media_type='',
        ),
        StreamFlavor(
            streams=[MockBackend('ffmpeg', 3)],
            bitrate=469,
            width=640,
            height=360,
            media_type='',
        ),
        failed_flavor('Second failure'),
        StreamFlavor(
            streams=[MockBackend('ffmpeg', 5)],
            bitrate=1506,
            width=1280,
            height=720,
            media_type='',
        ),
    ]

    assert backend_data(filter_flavors(test_flavors)) == [5]
    assert backend_data(
        filter_flavors(test_flavors, max_height=720, max_bitrate=200)
    ) == [2]
    assert backend_data(
        filter_flavors(test_flavors, max_height=400, max_bitrate=2000)
    ) == [3]


def test_combined_filter_bitrate_only_and_some_failures():
    test_flavors = [
        failed_flavor('Failure'),
        StreamFlavor(
            streams=[MockBackend('ffmpeg', 1)], bitrate=517, media_type='video'
        ),
        failed_flavor('Second failure'),
    ]

    assert backend_data(filter_flavors(test_flavors)) == [1]
    assert backend_data(filter_flavors(test_flavors, max_height=720)) == [1]
    assert backend_data(filter_flavors(test_flavors, max_bitrate=200)) == [1]
    assert backend_data(
        filter_flavors(test_flavors, max_height=720, max_bitrate=200)
    ) == [1]


def test_backend_filter_first_preferred():
    test_flavors = [
        video_flavor(
            [MockBackend('ffmpeg'), MockBackend('wget'), MockBackend('youtubedl')]
        )
    ]
    enabled = ['wget', 'ffmpeg', 'youtubedl', 'rtmpdump']
    flavor = filter_flavors(test_flavors, enabled_backends=enabled)

    assert flavor.streams[0].name == enabled[0]


def test_backend_filter_first_preferred_2():
    test_flavors = [video_flavor([MockBackend('rtmpdump')])]
    enabled = ['wget', 'ffmpeg', 'youtubedl', 'rtmpdump']
    flavor = filter_flavors(test_flavors, enabled_backends=enabled)

    assert flavor.streams[0].name == enabled[3]


def test_backend_filter_no_match():
    test_flavors = [
        video_flavor(
            [MockBackend('ffmpeg'), MockBackend('wget'), MockBackend('youtubedl')]
        )
    ]
    enabled = ['rtmpdump']
    flavor = filter_flavors(test_flavors, enabled_backends=enabled)

    assert len(flavor.streams) == 1
    assert not flavor.streams[0].is_valid()
    assert 'Required backend not enabled' in flavor.streams[0].error_message


def test_backend_filter_no_streams():
    flavor = filter_flavors([], enabled_backends=['ffmpeg'])

    assert flavor is None


def test_backend_filter_failed_stream():
    test_flavors = [
        video_flavor(
            [
                MockBackend('ffmpeg'),
                FailingBackend('wget stream failed'),
                MockBackend('youtubedl'),
            ]
        )
    ]
    enabled = ['wget']
    flavor = filter_flavors(test_flavors, enabled_backends=enabled)

    assert len(flavor.streams) == 1
    assert not flavor.streams[0].is_valid()


def test_backend_filter_failed_fallback():
    test_flavors = [
        video_flavor(
            [
                MockBackend('ffmpeg'),
                FailingBackend('wget stream failed'),
                MockBackend('youtubedl'),
            ]
        )
    ]
    enabled = ['wget', 'youtubedl', 'ffmpeg']
    flavor = filter_flavors(test_flavors, enabled_backends=enabled)

    assert len(flavor.streams) == 2
    assert flavor.streams[0].is_valid()
    assert flavor.streams[0].name == enabled[1]
    assert flavor.streams[1].is_valid()
    assert flavor.streams[1].name == enabled[2]



# This file is part of yle-dl.
#
# Copyright 2010-2022 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from datetime import datetime, timedelta
from yledl.timestamp import parse_areena_timestamp


def test_timestamp():
    ts = parse_areena_timestamp('2018-01-02T18:30:00+02:00')

    assert ts.replace(tzinfo=None) == datetime(2018, 1, 2, 18, 30, 00)
    assert ts.utcoffset() == timedelta(hours=2)


def test_timestamp_strip():
    ts = parse_areena_timestamp('  2018-01-02T18:30:00+02:00  ')
    assert ts.replace(tzinfo=None) == datetime(2018, 1, 2, 18, 30, 00)


def test_invalid_timestamp():
    assert parse_areena_timestamp('xxx2018-01-02T18:30:00+02:00') is None
    assert parse_areena_timestamp('2018-01-02T18:30:00') is None
    assert parse_areena_timestamp('2018-01-999999T18:30:00+02:00') is None
    assert parse_areena_timestamp('2018-01-999999T22222') is None



# This file is part of yle-dl.
#
# Copyright 2010-2025 Antti Ajanki and others
#
# Yle-dl is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Yle-dl is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with yle-dl. If not, see <https://www.gnu.org/licenses/>.

from utils import fetch_title, fetch_stream_url, fetch_metadata


def test_arkivet_title():
    title = fetch_title(
        'https://svenska.yle.fi/artikel/2014/06/13/halla-det-ar-naturvaktarna'
    )

    assert title
    assert title[0].startswith('Seportage om Naturväktarna')


def test_arkivet_stream_url():
    streamurl = fetch_stream_url(
        'https://svenska.yle.fi/artikel/2014/06/13/halla-det-ar-naturvaktarna'
    )
    assert streamurl
    assert '.m3u8' in streamurl[0]


def test_arkivet_metadata():
    metadata = fetch_metadata(
        'https://svenska.yle.fi/artikel/2014/06/13/halla-det-ar-naturvaktarna'
    )

    assert len(metadata) == 1
    assert metadata[0].get('title').startswith('Seportage om Naturväktarna')
    flavors = metadata[0]['flavors']
    assert any(f.get('media_type') == 'video' for f in flavors)
    assert all('bitrate' in f for f in flavors)
    assert any('height' in f for f in flavors)
    assert any('width' in f for f in flavors)


def test_arkivet_audio_stream_url():
    streamurl = fetch_stream_url(
        'https://svenska.yle.fi/artikel/2014/01/28/tove-jansson-laser-noveller-ur-dockskapet'
    )

    assert len(streamurl) == 11
    for url in streamurl:
        assert '.mp3' in url


def test_arkivet_audio_metadata():
    metadata = fetch_metadata(
        'https://svenska.yle.fi/artikel/2014/01/28/tove-jansson-laser-noveller-ur-dockskapet'
    )

    assert len(metadata) == 11
    assert metadata[0].get('title').startswith('Apan ur Dockskåpet')
    for m in metadata:
        assert all(f.get('media_type') == 'audio' for f in m.get('flavors'))


def test_arkivet_a__stream_url():
    streamurl = fetch_stream_url('https://svenska.yle.fi/a/7-884297')
    assert streamurl
    for url in streamurl:
        assert '/manifest.mpd' in url or '.m3u8' in url or '.mp3' in url


def test_arkivet_a_metadata():
    metadata = fetch_metadata('https://svenska.yle.fi/a/7-884297')

    assert len(metadata) >= 4
    assert metadata[1].get('title').startswith('Valborg på Borgbacken')
    flavors = metadata[1]['flavors']
    assert any(f.get('media_type') == 'video' for f in flavors)
    assert all('bitrate' in f for f in flavors)
    assert any('height' in f for f in flavors)
    assert any('width' in f for f in flavors)
